import {IAtom} from "./atom";
import {IDerivation} from "./derivation";
import {Reaction} from "./reaction";
import {invariant} from "../utils/utils";

declare const global: any;

/**
 * These values will persist if global state is reset
 */
const persistentKeys = ["mobxGuid", "resetId", "spyListeners", "strictMode"];

export class MobXGlobals {
	/**
	 * MobXGlobals version.
	 * MobX compatiblity with other versions loaded in memory as long as this version matches.
	 * It indicates that the global state still stores similar information
	 */
	version = 2;

	/**
	 * Stack of currently running derivations
	 */
	derivationStack: IDerivation[] = [];

	/**
	 * 'guid' for general purpose. Mostly debugging.
	 */
	mobxGuid = 0;

	/**
	 * Are we in a transaction block? (and how many of them)
	 */
	inTransaction = 0;

	/**
	 * Are we in an (un)tracked block?
	 */
	isTracking: boolean = false;

	/**
	 * Are we currently running reactions?
	 * Reactions are run after derivations using a trampoline.
	 */
	isRunningReactions = false;

	/**
	 * List of observables that have changed in a transaction.
	 * After completing the transaction(s) these atoms will notify their observers.
	 */
	changedAtoms: IAtom[] = [];

	/**
	 * List of scheduled, not yet executed, reactions.
	 */
	pendingReactions: Reaction[] = [];

	/**
	 * Is it allowed to change observables at this point?
	 * In general, MobX doesn't allow that when running computations and React.render.
	 * To ensure that those functions stay pure.
	 */
	allowStateChanges = true;
	/**
	 * If strict mode is enabled, state changes are by default not allowed
	 */
	strictMode = false;

	/**
	 * Used by createTransformer to detect that the global state has been reset.
	 */
	resetId = 0;

	/**
	 * Spy callbacks
	 */
	spyListeners: {(change: any): void}[] = [];
}

export const globalState: MobXGlobals = (() => {
	const res = new MobXGlobals();
	/**
	 * Backward compatibility check
	 */
	if (global.__mobservableTrackingStack || global.__mobservableViewStack)
		throw new Error("[mobx] An incompatible version of mobservable is already loaded.");
	if (global.__mobxGlobal && global.__mobxGlobal.version !== res.version)
		throw new Error("[mobx] An incompatible version of mobx is already loaded.");
	if (global.__mobxGlobal)
		return global.__mobxGlobal;
	return global.__mobxGlobal = res;
})();

export function registerGlobals() {
	// no-op to make explicit why this file is loaded
}

/**
 * For testing purposes only; this will break the internal state of existing observables,
 * but can be used to get back at a stable state after throwing errors
 */
export function resetGlobalState() {
	globalState.resetId++;
	const defaultGlobals = new MobXGlobals();
	for (let key in defaultGlobals)
		if (persistentKeys.indexOf(key) === -1)
			globalState[key] = defaultGlobals[key];
	globalState.allowStateChanges = !globalState.strictMode;
}
